/*
 *  Copyright (C) 2010 Nasser Giacaman, Oliver Sinnen
 *
 *  This file is part of Parallel Task.
 *
 *  Parallel Task is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or (at 
 *  your option) any later version.
 *
 *  Parallel Task is distributed in the hope that it will be useful, 
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General 
 *  Public License for more details.
 *  
 *  You should have received a copy of the GNU General Public License along 
 *  with Parallel Task. If not, see <http://www.gnu.org/licenses/>.
 */

package paratask.runtime;

import java.lang.reflect.Method;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.concurrent.BlockingQueue;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.ConcurrentLinkedQueue;
import java.util.concurrent.CountDownLatch;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.atomic.AtomicBoolean;
import java.util.concurrent.atomic.AtomicInteger;
import java.util.concurrent.locks.ReentrantLock;


import paratask.queues.PipelineQueue;

/**
 * A future object representing a task invocation. As well as containing the return result for non-<code>void</code> tasks, 
 * the <code>TaskID</code> may also be used with <code>dependsOn</code>, cancel attempts, and other various functions. 
 * 
 * @author Nasser Giacaman
 * @author Oliver Sinnen
 *
 * @param <E> The task's return type
 */
public class TaskID<E> {
	
	static protected AtomicInteger nextGlobalID = new AtomicInteger(-1);
	
	protected int globalID = -1;
	protected int relativeID = 0;
	
	protected TaskID<?> enclosingTask = null;	// this is used in case we need to find an asynchrous exception handler
	
	private int executeOnThread = ParaTaskHelper.ANY_THREAD_TASK;
	
	protected TaskInfo taskInfo = null;
	private E returnResult = null;
	
	private int progress = 0;

	protected AtomicBoolean hasCompleted = null;
	protected boolean cancelled = false;
	protected AtomicBoolean cancelRequested = new AtomicBoolean(false);
	
    //-- the registering thread has its own latch, since it is allowed to progress inside slots of this TaskID
    private CountDownLatch completedLatchForRegisteringThread= null;
    
    //-- all the other threads (non-registering threads) must wait at this latch, until slots complete
    private CountDownLatch completedLatch = null;
    
	private ReentrantLock changeStatusLock = null;
	
	protected AtomicBoolean hasUserError = new AtomicBoolean(false);
	private Throwable exception = null;
	
	private boolean isInteractive = false;
	
	private ConcurrentLinkedQueue<TaskID> waitingTasks = null;	//-- TaskIDs waiting for this task 
	private ConcurrentHashMap<TaskID, Object> remainingDependences = null;	//-- TaskIDs this task is waiting for
	
	protected TaskIDGroup group = null;
	
	protected boolean hasSlots = false;

	static final protected int CREATED = 0;
	static final protected int CANCELLED = 1;
	static final protected int STARTED = 2;
	protected AtomicInteger status = new AtomicInteger(CREATED);
	
	// pipeline stuff
	private List<PipelineQueue<E>> outputQueues = new ArrayList<PipelineQueue<E>>();
	private boolean firstQueueClaimed = false;
	private PipelineThread pipelineThread = null; 
	
	/**
	 * Checks to see if this task has successfully cancelled.
	 * @return <code>true</code> if it has cancelled successfully, <code>false</code> otherwise. 
	 */
	public boolean cancelledSuccessfully() {
		return cancelled;
	}
	
	/**
	 * Checks to see if this task is an interactibe task.
	 * 
	 * @return <code>true</code> if this is an interactive task, <code>false</code> otherwise.
	 */
	public boolean isInteractive() {
		return isInteractive;
	}
	
	/**
	 * Checks to see if this task if a part of a pipeline.
	 */
	public boolean isPipeline() {
		return taskInfo.isPipeline();
	}
	
	/**
	 * Assigns the thread for this stage of the pipeline. Used for cancelling
	 * the stage via PipelineThread.cancel().
	 */
	protected void setPipelineThread(PipelineThread pt) {
		if (!isPipeline()) 
			throw new IllegalStateException("trying to assign PipelineThread to non-pipeline task");
		
		this.pipelineThread = pt;
	}
	
	/**
	 * If this is a pipeline stage, get a queue from which all future results
	 * can be retrieved. If not, returns null.
	 * 
	 * Thread-safe.
	 */
	public BlockingQueue<E> getOutputQueue() {
		return getOutputQueue(null);
	}
	
	/**
	 * If this is a pipeline stage, get a queue from which all future results
	 * can be retrieved. If not, returns null.
	 * 
	 * Intended for internal use only. Must pass the TaskID of the task
	 * requesting the queue in order to associate the queue with the tasks it
	 * connects.
	 * 
	 * Thread-safe.
	 */
	protected BlockingQueue<E> getOutputQueue(TaskID requester) {
		if (!isPipeline())
			return null;
		
		synchronized(outputQueues) {
			if (!firstQueueClaimed) {
				firstQueueClaimed = true;
				if (outputQueues.size() == 0) {
					PipelineQueue<E> queue = new PipelineQueue<E>(this, requester);
					outputQueues.add(queue);
					return queue;
				} else {
					PipelineQueue<E> queue = outputQueues.get(0);
					queue.setTailTask(requester);
					return queue;
				}
			} else {
				PipelineQueue<E> queue = new PipelineQueue<E>(this, requester);
				outputQueues.add(queue);
				return queue;
			}
		}
	}
	
	/**
	 * Unregisters a queue from a pipeline stage so that new results are no
	 * longer written to it. Also, releases the reference so it can be GC'd.
	 * 
	 * For internal use only.
	 * 
	 * Thread-safe.
	 */
	protected void unregisterOutputQueue(PipelineQueue<E> queue) {
		if (!isPipeline()) {
			// maybe replace this with "return;"
			throw new IllegalStateException("trying to remove a queue from non-pipeline");
		}
		
		synchronized(outputQueues) {
			if (!outputQueues.contains(queue))
				throw new IllegalArgumentException(pipelineThread.getName() + ": queue to unregister not in outputQueues");
			
			outputQueues.remove(queue);
			
			// prevent this from being called again
			queue.setHeadTask(null);
			
			// reset firstQueueClaimed flag
			if (outputQueues.size() == 0)
				firstQueueClaimed = false;
		}
	}
	
	/**
	 * Write something into the output queue(s) handled by this TaskID. Only
	 * existing queues will get the object.
	 * 
	 * Thread-safe.
	 */
	protected void writeToOutputQueues(E value) {
		if (!isPipeline()) {
			// maybe replace this with "return;"
			throw new IllegalStateException("trying to write to output queue when not a pipeline");
		}
		
		synchronized(outputQueues) {
			if (outputQueues.size() == 0) {
				outputQueues.add(new PipelineQueue<E>(this, null));
			}
			
			for (PipelineQueue<E> queue : outputQueues) {
				queue.add(value);
			}
		}
	}
	
	/**
	 * For internal use only. Cancel all child stages if this is a pipeline.
	 */
	protected void cancelChildTasks() {
		if (!isPipeline()) {
			// maybe replace this with "return;"
			throw new IllegalStateException("trying to cancel child stages when not a pipeline");
		}
		
		synchronized(outputQueues) {
			for (PipelineQueue<E> queue : outputQueues) {
				if (queue.getTailTask() != null)
					queue.getTailTask().cancelAttempt();
			}
		}
	}
	
	/**
	 * Sets the progress of the task.
	 * @param progress	The progress to set for the task
	 * @see CurrentTask#getProgress()
	 * @see CurrentTask#setProgress(int)
	 * @see #getProgress()
	 */
	void setProgress(int progress) {
		this.progress = progress;
	}

	/**
	 * Returns the current progress of this task
	 * @return	The task's current progress.
	 * @see CurrentTask#getProgress()
	 * @see CurrentTask#setProgress(int)
	 * @see #setProgress(int)
	 * 
	 */
	public int getProgress() {
		return progress;
	}
	
	/**
	 * Checks to see if this task has been requested to cancel.
	 * @return <code>true</code> if it has been requested to cancel, <code>false</code> otherwise.
	 * @see CurrentTask#cancelRequested()
	 * @see #cancelAttempt()
	 * @see #cancelledSuccessfully()
	 */
	public boolean cancelRequested() {
		return cancelRequested.get();
	}
	
	TaskID(boolean alreadyCompleted) {
		if (alreadyCompleted) {
			globalID = nextGlobalID.incrementAndGet();
			completedLatch = new CountDownLatch(0);
			completedLatch = new CountDownLatch(0);
			hasCompleted = new AtomicBoolean(true);
			status = new AtomicInteger(STARTED);
		} else {
			throw new UnsupportedOperationException("Don't call this constructor if passing in 'false'!");
		}
	}
	
	TaskID() {
		globalID = nextGlobalID.incrementAndGet();
		completedLatch = new CountDownLatch(1);
		hasCompleted = new AtomicBoolean(false);
		status = new AtomicInteger(CREATED);
		changeStatusLock = new ReentrantLock();
	}
	
	TaskID(TaskInfo taskInfo) {
		this();
		completedLatchForRegisteringThread = new CountDownLatch(1);
		this.taskInfo = taskInfo;
		isInteractive = taskInfo.isInteractive();
		if (taskInfo != null) {
			hasSlots = taskInfo.getSlotsToNotify() != null;
//			hasHandlers = taskInfo.hasRegisteredHandlers();
		}
	}
	
	//-- return true if successfully canceled
	//-- TODO at the moment this is not a perfect implementation, since a 2nd (and 3rd, etc) call to this method might return false, 
	// even if it has successfully cancelled.. (this can be correctly by locking - but don't want to introcuce the overhead of locking every 
	// time a task is started..)
	/**
	 * Attempts to cancel the task. If cancelled successfully, the task will not be enqueued. A failed cancel 
	 * will still allow the task to continue executing. To stop the task, the task should check
	 * to see if a cancel request has been made. 
	 * @return <code>true</code> if it has cancelled successfully, <code>false</code> otherwise.
	 * @see #cancelRequested() 
	 * @see CurrentTask#cancelRequested()
	 * @see #cancelledSuccessfully()
	 */
	public boolean cancelAttempt() {
		cancelRequested.set(true);
		
		if (isPipeline()) {
			// this tells PipelineThread that some parent has requested cancel
			// it can be called more than once because of multiple parent stages
			pipelineThread.cancel();
			cancelled = true;
		}
		
		int prevStatus = status.getAndSet(CANCELLED);
		
		// TODO currently this only returns true if this is the first time attempting to cancel.. should be 
		// fixed so that it also returns true if it was cancelled before..(but this requires locking etc).
		
		if (prevStatus == CREATED || cancelled) {
			cancelled = true;
			return true;
		}
		return false;
	}
	
	/**
	 * Returns the group that this task is part of (assuming it is a multi-task).  
	 * @return	Returns the group associated with this task. If not part of a multi-task, then
	 * returns <code>null</code>.
	 */
	public TaskIDGroup getGroup() {
		return group;
	}
	
	boolean executeAttempt() {
		int prevStatus = status.getAndSet(STARTED);
		
		//-- TODO if the status was not CREATED before this, then should probably set status to CANCELLED 
		return prevStatus == CREATED;
	}
	
	void setEnclosingTask(TaskID enclosingTask) {
		this.enclosingTask = enclosingTask;
	}
	
	TaskID getEnclosingTask() {
		return enclosingTask;
	}
	
	void addWaiter(TaskID waiter) {
		if (hasCompleted.get()) {
			waiter.dependenceFinished(this);
		} else {
			changeStatusLock.lock();
			
			if (!hasCompleted.get()) {
				if (waitingTasks == null)
					waitingTasks = new ConcurrentLinkedQueue<TaskID>();
				
				waitingTasks.add(waiter);
			} else {
				waiter.dependenceFinished(this);
			}
			changeStatusLock.unlock();
		}
	}
	
	/**
	 * Returns the (method) name of the task.
	 * @return
	 */
	public String getTaskName() {
		return taskInfo.getMethod().getName().replaceFirst(ParaTaskHelper.PT_PREFIX, "");
	}
	
	/* one of the other Tasks (that this task dependsOn) has finished 
	 * 
	 * if that was the last Task we were waiting for, inform the taskpool
	 * 
	 * */
	void dependenceFinished(TaskID otherTask) {
		remainingDependences.remove(otherTask);
		if (remainingDependences.isEmpty()) {
			TaskpoolFactory.getTaskpool().nowReady(this);
		}
	}
	
	void setRemainingDependences(ArrayList<TaskID> deps) {
		remainingDependences = new ConcurrentHashMap<TaskID, Object>();
		Iterator<TaskID> it = deps.iterator();
		while (it.hasNext()) {
			remainingDependences.put(it.next(), "");
		}
	}
	
	/**
	 * Returns the task's globally-unique ID.
	 * @return	The task's unique ID.
	 * @see CurrentTask#globalID()
	 * @see CurrentTask#relativeID()
	 * @see #relativeID()
	 */
	public int globalID() {
		return globalID;
	}
	
	/**
	 * Returns the sub-task's relative ID in the multi-task. 
	 * @return	The position, starting from 0, of this sub-task compared to it's sibling subtasks.
	 * @see CurrentTask#globalID()
	 * @see CurrentTask#relativeID()
	 * @see #globalID()  
	 */
	public int relativeID() {
		return relativeID;
	}
	
	void setRelativeID(int relativeID) {
		this.relativeID = relativeID;
	}
	
	TaskInfo getTaskInfo() {
		return taskInfo;
	}
	
	/**
	 * Returns the result of the task. If the task has not finished yet, the current thread blocks. 
	 * ParaTask worker threads will not block, instead they execute other ready tasks until
	 * this task completes.
	 * @return	The result of the task.
	 * @throws ExecutionException
	 * @throws InterruptedException
	 */
	public E getReturnResult() throws ExecutionException, InterruptedException {
		waitTillFinished();
		if (cancelledSuccessfully())
			throw new ParaTaskRuntimeException("Attempting to get the result of a cancelled Task!");
		return returnResult;
	}
	
	/**
	 * Blocks the current thread until the task finishes. If the blocking thread is a 
	 * ParaTask worker thread, then other tasks are executed until this task completes. 
	 * @throws ExecutionException
	 * @throws InterruptedException
	 */
	public void waitTillFinished() throws ExecutionException, InterruptedException {
		if (!hasCompleted.get()) {
			Thread t = Thread.currentThread();
			
			/* Only WorkerThreads should start a new TaskID.. all other threads belong to the user, or 
			 * are InteractiveThreads (therefore it is OK for them to block) */
			if (t instanceof WorkerThread) {
				WorkerThread currentWorker = (WorkerThread) t;
				
				// if still not completed then start a substitute thread
				while (!hasCompleted.get()) {
					
					// causes the worker to either execute a task or sleep
					currentWorker.executeAnotherTaskOrSleep();
				}
			} else {
				if (currenThreadIsTheRegisteredThread()) {
//					System.out.println("current thread is the registering trhead.. will await");
					completedLatchForRegisteringThread.await();
//					System.out.println("finsehd await");
				} else {
					completedLatch.await();
				}
			}
		}
		
		//-- task has completed.. was there a user error?
		if (hasUserError.get()) {
			throw new ExecutionException(exception);
		}
	}
	
	protected boolean currenThreadIsTheRegisteredThread() {
		Thread registered = taskInfo.getRegisteringThread();
		if (registered == null)
			return false;
		if (registered == EventLoop.EDT && GuiThread.isEventDispatchThread())
			return true;
		else 
			return registered == Thread.currentThread();
	}
	
	void setReturnResult(E returnResult) {
		this.returnResult = returnResult;
	}
	
	/**
	 * Returns the exception that occurred while this task executed. 
	 * @return	The exception that occurred.
	 */
	public Throwable getException() {
		return exception;
	}
	
	void setException(Throwable exception) {
		this.exception = exception;
		hasUserError.set(true);
		if (group != null) {
			group.setException(exception);
		}
	}
	
	/**
	 * Checks to see whether the task has completed.
	 * @return	<code>true</code> if it has completed, <code>false</code> otherwise
	 * @see #getProgress()
	 * @see CurrentTask#getProgress()
	 */
	public boolean hasCompleted() {
		return hasCompleted.get();
	}

	/**
	 * Checks to see whether the task had any errors. 
	 * @return	<code>true</code> if there was an error, <code>false</code> otherwise
	 */
	public boolean hasUserError() {
		return hasUserError.get();
	}
	
	/*
	 * The Task is really complete now, since any slots have been executed by the registering thread.
	 * 
	 * This method will be called by the registering thread when it completes all the slots (handlers and notifies)
	 */
	void setComplete() {
		
//		System.out.println("SET COMPLETE...");
		
		/* dependences (gotta be atomic) */
		changeStatusLock.lock();
		TaskID waiter = null;
		
		if (waitingTasks != null) {
			while ((waiter = waitingTasks.poll()) != null) {
				// removes the waiter from the queue
				waiter.dependenceFinished(this);
			}
		}
		
		completedLatchForRegisteringThread.countDown();	//-- in case there were no slots
		completedLatch.countDown();
		hasCompleted.set(true);
		changeStatusLock.unlock();
		
		if (group != null) {
			group.setComplete();
		}
	}
	
	/*
	 *	This will enqueue the notify slots and the exception handlers to be executed by the registered thread. 
	 *	Note that this TaskID is NOT considered completed, until all these slots are finished (even though the 
	 *	actual task logic has been executed, we need to wait for the slots before handling dependences, etc).
	 *
	 * 	Therefore, the registering thread will later set the status of this task as complete. 
	 */
	void enqueueSlots(boolean onlyEnqueueFinishedSlot) {
		
		// TODO make use of the boolean passed in (and for multi-tasks) - i.e. don't execute slots of cancelled tasks?
		
		if (group != null && group.isMultiTask()) {
			//-- part of a multi-task, will only enqueue the slots of the group when the last TaskID in the group completes
			group.oneMoreInnerTaskCompleted();
			setComplete();	// TODO  this was never here before -- was deadlocking without it... 20/3/2010 
			
		} else {
			//-- even if this TaskID is within a group, it is a separate entity since not a multi-task
			
			// TODO   if it has been cancelled, then probably don't want to execute the handlers... (except the setCompleteMethod())
			if (hasUserError() || hasSlots) {
				
//				System.out.println("HAS SLOTS..");
				completedLatchForRegisteringThread.countDown(); //-- so that registering thread will not block in slots 
				completedLatch.countDown();  // ADDED THIS??
				
				if (hasUserError.get())
					executeHandlers();
				if (hasSlots)
					executeSlots();
				
				//-- 		since slots are executed in the order they are enqueued, then this will be the last slot! :-)
				callTaskListener(new Slot(ParaTaskHelper.setCompleteSlot, this, false, false, true));
				
			} else {
				setComplete();
			}
		}
	}
	
	protected Slot getExceptionHandler(Class occurredException) {
		
		//-- first, try to get handler defined immediately for this task
		Slot handler = taskInfo.getExceptionHandler(occurredException);
		
		TaskID curTask = this;
		//-- while we have not found a handler, and while there are other enclosing tasks
		while (handler == null && curTask.getEnclosingTask() != null) {
			curTask = curTask.getEnclosingTask();
			handler = curTask.getTaskInfo().getExceptionHandler(occurredException);
		}
		
		return handler;
	}
	
	//-- returns the number of handlers that it will execute for this TaskID
	private int executeHandlers() {
		Slot handler = getExceptionHandler(exception.getClass());
		
		if (handler != null) {
			callTaskListener(handler);
			return 1;
		} else {
			String taskName = "";
			Method method = taskInfo.getMethod();
			if (method.getDeclaringClass().getPackage() != null)
				taskName+=method.getDeclaringClass().getPackage().getName()+".";
			taskName+=getTaskName()+"(";
			Class[] params = method.getParameterTypes();
			if (params != null) {
				for (int p = 0; p < params.length; p++) {
					taskName+=params[p].getName();
					if (p != (params.length-1))
						taskName+=", ";
				}
			}
			taskName+=")";
			System.err.println("No asynchronous exception handler (i.e. asyncCatch clause) was specified when " +
					"invoking:\n\t\t "+taskName +" " +
							"\n\tThe globalID of this task is "+ globalID() + ", and the encountered exception: ");
			exception.printStackTrace();
			return 0;
		}
	}
	
	void callTaskListener(Slot slot) {
		Thread registeredThread = taskInfo.getRegisteringThread();
		
//		System.out.println("want to execute slot: "+slot.getMethod().getName());
		
		TaskListener listener = null;
		if (slot.executeOnEDT()) {
			listener = EventLoop.getEDTTaskListener();
		} else if (registeredThread != null) {
			listener = EventLoop.getTaskListener(registeredThread);
		}
		
		// TODO  if the registering thread is a TaskThread, then execute on the slot handling thread
		
		if (listener != null) {
			listener.executeSlots(slot);
		} else {
			if (registeredThread != null) {
				if (ParaTaskHelper.isSubClassOf(registeredThread.getClass(),TaskThread.class)) {
					//-- the registering thread was a TaskThread, therefore the slot is executed sequentially by the special SlotHandlingThread
					EventLoop.executeOnSlotHandlingThread(slot);
				} else {
					System.err.println("Tried to execute slot (TaskID::callTaskListener()). But There is no task listener for thread " + registeredThread);
					// if can't find listener, then never reach here... unless i've missed something out
					throw new ParaTaskRuntimeException("can't find listener");
				}
			} else {
				// TODO there is no registered thread... 
			}
		}
//		System.out.println("executed slot: "+slot.getMethod().getName());
	}
	
	protected int executeIntermediateSlots() {
		return 0;
	}
	
	protected int executeSlots() {
		for (Iterator<Slot> it = taskInfo.getSlotsToNotify().iterator(); it.hasNext(); ) 
			callTaskListener(it.next());
		return taskInfo.getSlotsToNotify().size();
	}
	
	/**
	 * Returns the arguments that were passed to this task when it was invoked.  
	 * 
	 * @return The arguments initially passed to the task when it was invoked
	 */
	public Object[] getTaskArguments() {
		return taskInfo.getParameters();
	}
	
	void setPartOfGroup(TaskIDGroup group) {
		this.group = group;
	}
	
	/**
	 * Checks to see if this task is part of a multi-task group.
	 * @return <code>true</code> if this task is part of a multi-task, <code>false</code> otherwise 
	 */
	public boolean isMultiTask() {
		return group != null;
	}
	
	/**
	 * Returns the size of the multi-task this task is part of. 
	 * @return	The multi-task size, otherwise returns 1 if this task is not part of a multi-task.
	 */
	public int multiTaskSize() {
		if (group == null)
			return 1;
		return group.groupSize();
	}
	
	int getExecuteOnThread() {
		return executeOnThread;
	}
	
	void setExecuteOnThread(int executeOnThread) {
		this.executeOnThread = executeOnThread;
	}
}
